/************************************************************************
 *									*
 *            DOS.C							*
 *            =====							*
 *                                        W.G. Reissnegger		*
 *                                        C&C Research Labs		*
 *                                        NEC USA Inc.			*
 *                                        4 Independence Way		*
 *                                        Princeton NJ 08540		*
 *									*
 *   DOS FAT12 and FAT16 messages                                       *
 *                                                                      *
 *   Copyright (C) 1997, 1998, 1999, 2000 NEC USA, Inc.			*
 *                 All Rights Reserved					*
 *									*
 *   This program is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License as     *
 *   published by the Free Software Foundation; either version 2 of     *
 *   the license, or (at your option) any later version.                *
 *									*
 *   This program is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of     *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      *
 *   GNU General Public License for more details.                       *
 *                                                                      *
 *   You should have received a copy of the GNU General Public          *
 *   License along with this program;  if not, write to the             *
 *       Free Software Foundation, Inc.,                                *
 *       59 Temple Place - Suite 330,                                   *
 *       Boston, MA 02111-1307                                          *
 *                                                                      *
 ************************************************************************/

#include <Messages.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stream.h>
#include "msdos.h"

#define D(c)
#define DOS_MEM_DEBUG			0xD0
#define DOS_DEFAULT_MBLK_SIZE		512

static DOS_INFO dos;

/************************************************************************
 *                                                                      *
 *            Downstream DOS I/O Routines                               *
 *            ===========================                               *
 *                                                                      *
 ************************************************************************/

void dos_fs_read_cluster(FILE *fp, MSDOS_FS_INFO *inf,
			 ushort cluster, ROME_MESSAGE *msg)
{
    msg->b_immed = inf->cluster0_lba + cluster * inf->cluster_size;
    msg->b_wptr = (uchar *)(inf->cluster_size * inf->sector_size);
    rome_getmblk(fp, msg);
    rome_await_message(msg, 0);
}

void dos_fs_read_dir_cluster(FILE *fp, MSDOS_FS_INFO *inf,
			     ushort cluster, ROME_MESSAGE *msg)
{
    msg->b_immed = inf->root_dir_lba + cluster * inf->cluster_size;
    msg->b_wptr = (uchar *)(inf->cluster_size * inf->sector_size);
    rome_getmblk(fp, msg);
    rome_await_message(msg, 0);
}

void dos_fs_write_cluster(FILE *fp, MSDOS_FS_INFO *inf,
			  ushort cluster, ROME_MESSAGE *msg)
{
    msg->b_immed = inf->cluster0_lba + cluster * inf->cluster_size;
    rome_putmblk(fp, msg);
    rome_await_message(msg, 0);
}

/************************************************************************
 *									*
 *            Support Routines						*
 *            ================						*
 *									*
 ************************************************************************/

static int read_partition_table(FILE *fp, MSDOS_FS_INFO *inf)
{
    uchar known_ptypes[] =
    {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a,
	0x0b, 0x40, 0x51, 0x52, 0x63, 0x64, 0x65, 0x75, 0x80, 0x81, 0x82,
	0x83, 0x93, 0x94, 0xa5, 0xb7, 0xb8, 0xc7, 0xdb, 0xe1, 0xe3, 0xf2,
	0xff
    };
    int bad_partition_table;
    int partition_entry_found;
    ROME_MESSAGE msg;
    MSDOS_PARTITION_TABLE p[4];
    int i;
    int active = -1;

    /* Three possibilities:
       1.  no disk at all, or not a DOS disk -- return 1;
       2.  not a partitioned disk -- return 0 with bs_lba = -1;
       3.  found a boot partition -- return 0 with bs_lba pointing to start
    */

    inf->bs_lba = -1;

    msg.b_immed = 0;
    msg.b_wptr = (uchar *)512;
    rome_getmblk(fp, &msg);
    rome_await_message(&msg, 0);

    if (msg.m_errno != 0)
    {
	return 1;
    }

    memcpy(p, msg.b_rptr + 446, 64);

    if (msg.b_rptr[510] != 0x55 || msg.b_rptr[511] != 0xaa)
    {
	rome_retmblk(&msg);
	return 1;
    }

    rome_retmblk(&msg);

    partition_entry_found = FALSE;
    bad_partition_table = FALSE;

    for (i = 0; i < 4; i++)
    {
	if ((p[i].active & 0x7f) == 0)
	{
	    int j;

	    for (j = 0; sizeof(known_ptypes) > j; j++)
	    {
		if (p[i].type == known_ptypes[j])
		{
		    break;
		}
	    }
	    if (j == sizeof(known_ptypes))
	    {
		return 0;
	    }
	    if (0x80 == p[i].active)
	    {
		active = i;
	    }
	    if (p[i].type)
	    {
		partition_entry_found = TRUE;
	    }
	}
    }

    if (active != -1)
    {
	inf->bs_lba = p[active].start_lba;
    }
    else if (partition_entry_found)
    {
	inf->bs_lba = p[0].start_lba;
    }

    return 0;
}

static int dos_read_fs_info(const char *proc, FILE * fp, MSDOS_FS_INFO * inf)
{
    ROME_MESSAGE msg;
    MSDOS_BOOT_SECTOR *bs;

    if  (read_partition_table(fp, inf) != 0)
    {
	return 1;
    }

    if (inf->bs_lba == -1)
    {
	/* assume an unpartitioned disk */

	inf->bs_lba = 0;
    }

    /*
     * read the boot sector
     */

    msg.b_immed = inf->bs_lba;
    msg.b_wptr = (uchar *) 512;
    rome_getmblk(fp, &msg);
    rome_await_message(&msg, 0);
    bs = (MSDOS_BOOT_SECTOR *)msg.b_rptr;

    /* this only works for MSDOS 5.0 or higher formatted disks */

    if ('6' == bs->file_id[4])
    {
	inf->fat_type = DOS_FAT16;
    }
    else if ('2' == bs->file_id[4])
    {
	inf->fat_type = DOS_FAT12;
    }
    else
    {
	int bits = 512*bs->fat_length*8;
	int sects = bs->sectors[0] | (bs->sectors[1] << 8);
	int bps = bits/sects;

	if  (bps == 12)
	{
	    inf->fat_type = DOS_FAT12;
	}
	else if  (bps == 16)
	{
	    inf->fat_type = DOS_FAT16;
	}
	else
	{
	    rome_retmblk(&msg);
	    return 1;
	}
    }

    inf->totalsize = inf->sector_size * bs->total_sect ?
	bs->total_sect : bs->sectors[0] | bs->sectors[1];

    /* not so very nice, but it works */

    inf->fat_lba = inf->bs_lba + 1;
    inf->fat_length = bs->fat_length;

    inf->sector_size = bs->sector_size[0] | bs->sector_size[1] << 8;
    inf->cluster_size = bs->cluster_size;
    inf->blksize = inf->cluster_size * inf->sector_size;

    inf->root_dir_lba = inf->fat_lba + bs->fats * bs->fat_length;
    inf->root_dir_entries = bs->dir_entries[0] | bs->dir_entries[1] << 8;

    inf->cluster0_lba = inf->root_dir_lba - (2 * inf->cluster_size) +
	((inf->root_dir_entries * sizeof(MSDOS_DIR_ENTRY))
	 / inf->sector_size);

    rome_retmblk(&msg);

    return dos_fat_read(fp, inf);
}

static DOS_DEVICE *dos_setup(char *path, char *fn)
{
    char devid = *path;
    DOS_DEVICE *dev = dos.dev_list;

    /* split into device and filename */

    while (*path && ('/' != *path++)) ;

    strcpy(fn, path);

    while (dev)
    {
	if (devid == dev->devid)
	{
	    break;
	}

	dev = dev->next;
    }

    if  (dev)
    {
	if  (dev->inf_valid == 0)
	{
	    if (0 != dos_read_fs_info(dev->proc_name, dev->fp, &dev->inf))
	    {
		return NULL;
	    }
	    dev->inf_valid = 1;
	}
    }

    return dev;
}



/************************************************************************
 *									*
 *          Messages							*
 *	    ========							*
 *									*
 ************************************************************************/

/************************************************************************
 *									*
 *            Messages							*
 *            ========							*
 *									*
 ************************************************************************/

void dos_m_close(ROME_MESSAGE *mptr)
{
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    int rc;
    DOS_DEVICE *dev = of->dev;
    MSDOS_DIR_ENTRY entry;
    ROME_MESSAGE *rep;

    /* check if we have to write a pending buffer */

    if  (of-> dev && of->dirty)
    {
	if (0 != (rc = dos_dir_get_entry(dev->fp, &dev->inf,
					 of->filename, &entry)))
	{
	    rome_fatal("file_close(): No DIR entry\n");
	}

	/* we write the complete cluster, although there might be used
	 * only a few bytes of it.
	 * the size field in the dir entry reflects the actual size.
	 */

	of->mblk.b_wptr = of->mblk.b_rptr + of->dev->inf.blksize;
	dos_fs_write_cluster(dev->fp, &dev->inf, of->cluster, &of->mblk);

	/* set the EOF fat entry (although this entry should already be set) */

	if (0 != dos_fat_write_entry(&dev->inf, of->cluster, FAT_EOF))
	{
	    rome_kprintf("dos_close(): Could not terminate FAT chain for "
			 "file %s\n", of->filename);
	    mptr->m_errno = EIO;
	}

	/* close the file */

	entry.size = of->fend;
	entry.time = dos_dir_mktime();
	entry.date = dos_dir_mkdate();

	if (0 != (rc = dos_dir_write_entry(dev->fp, &dev->inf,
					   of->filename, &entry)))
	{
	    mptr->m_errno = rc;
	}
	else if (0 != dos_fat_write(dev->fp, &dev->inf))
	{
	    mptr->m_errno = EIO;
	}
    }

    /*  All other messages will be canceled */

    while (NULL != (rep = rome_remhead(&of->q)))
    {
	rep->m_errno = ECANCELED;
	rome_reply(rep);
    }

    if (dos.o_root == of)
    {
	dos.o_root = of->link;
    }
    else
    {
	DOS_OPENFILE *tmp = dos.o_root;

	while (tmp)
	{
	    if (tmp->link == of)
	    {
		tmp->link = of->link;
		break;
	    }
	    tmp = tmp->link;
	}

	if (tmp == NULL)
	{
	    rome_fatal("dos: bad close\n");
	}
    }
    rome_free(of);

    rome_reply(mptr);

}

void dos_m_command(ROME_MESSAGE *mptr)
{
    char *cmd = mptr->b_rptr;
    char dname[32];
    char devid;

    if (2 == sscanf(cmd, "drive %s id:%c", dname, &devid))
    {
	DOS_DEVICE *dev;
	FILE *fp;

	if (!(fp = fopen(dname, "rw")))
	{
	    rome_kprintf("DOS: Could not open '%s'\n", dname);
	}
	else
	{
	    ROME_MESSAGE *e;
	    int i;

	    dev = rome_alloc(sizeof(DOS_DEVICE), DOS_MEM_DEBUG, 1);
	    dev->fp = fp;
	    dev->devid = devid;
	    strcpy(dev->proc_name, dname);
	    dev->inf_valid = 0;

	    dev->next = dos.dev_list;
	    dos.dev_list = dev;

	    for  (i = 0; 2 > i; i++)
	    {
		e = (ROME_MESSAGE *)rome_alloc(sizeof(ROME_MESSAGE), DOS_MEM_DEBUG, 1);

		e->src_context = dev;
		rome_get_event(fp, e, rome_this_ptr->prio);
	    }
	}
    }
    else
    {
	rome_kprintf("dos_command(): Unknown DOS command '%s'\n", cmd);
    }

    rome_reply(mptr);

}

void dos_r_event(ROME_MESSAGE *mptr)
{
    ROME_P_MEDIA_MEDIUM_EJECT *ej = RCASTP(ROME_P_MEDIA_MEDIUM_EJECT, mptr);
    DOS_DEVICE *dev = (DOS_DEVICE *)mptr->src_context;

    if  (ej->event_type == ROME_E_MEDIA_MEDIUM_EJECT)
    {
	DOS_OPENFILE *xx = dos.o_root;

	while (xx)
	{
	    if (xx->dev == dev)
	    {
		xx->dev = NULL;
	    }
	}

	if  (dev->inf.fat)
	{
	    rome_free(dev->inf.fat);
	    dev->inf.fat = NULL;
	}
	dev->inf_valid = 0;
    }
    rome_get_event(dev->fp, mptr, rome_this_ptr->prio);
}

void dos_m_fslib_attr_get(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_ATTR_GET *p = RCASTP(ROME_T_FSLIB_ATTR_GET, mptr);
    int rc;
    char path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *dev;
    MSDOS_DIR_ENTRY entry;
    FSLIB_ATTR *attr = p->attr;

    if ((dev = dos_setup(p->name, path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else if (0 != (rc = dos_dir_get_entry(dev->fp, &dev->inf, path, &entry)))
    {
	mptr->m_errno = rc;
    }
    else
    {
	attr->type = ((ATTR_DIR & entry.attr) ? FSLIB_TYPE_DIR : FSLIB_TYPE_FILE);
	attr->size = entry.size;
    }

    rome_reply(mptr);

}

void dos_m_fslib_attr_set(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_ATTR_SET *p = RCASTP(ROME_T_FSLIB_ATTR_SET, mptr);
    int rc;
    char path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *dev;
    MSDOS_DIR_ENTRY entry;
    FSLIB_ATTR *attr = p->attr;

    if ((dev = dos_setup(p->name, path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else if (0 != (rc = dos_dir_get_entry(dev->fp, &dev->inf, path, &entry)))
    {
	mptr->m_errno = rc;
    }
    else
    {
	/* set attribute values here */

	entry.size = attr->size;

	if (FSLIB_TYPE_FILE == attr->type)
	{
	    entry.attr = ATTR_ARCH;
	    mptr->m_errno = dos_dir_write_entry(dev->fp, &dev->inf, path, &entry);
	}
	else if (FSLIB_TYPE_DIR == attr->type)
	{
	    entry.attr = ATTR_DIR;
	    mptr->m_errno = dos_dir_write_entry(dev->fp, &dev->inf, path, &entry);
	}
	else
	{
	    mptr->m_errno = EINVAL;
	}
    }
    rome_reply(mptr);

}

void dos_m_fslib_delete(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_DELETE *p = RCASTP(ROME_T_FSLIB_DELETE, mptr);
    char path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *dev;

    if ((dev = dos_setup(p->name, path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else if ((mptr->m_errno = dos_dir_rmfile(dev->fp, &dev->inf, path) == 0))
    {
	mptr->m_errno = dos_fat_write(dev->fp, &dev->inf);
    }

    rome_reply(mptr);

}

void dos_m_fslib_dev_attr(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_DEV_ATTR *p = RCASTP(ROME_T_FSLIB_DEV_ATTR, mptr);
    FSLIB_DEVICE_ATTR *attr = p->attr;
    char path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *dev;

    if ((dev = dos_setup(p->devspec, path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else
    {
	attr->maxblk = dev->inf.totalsize;
    }

    rome_reply(mptr);

}

void dos_m_fslib_dev_list(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_DEV_LIST *p = RCASTP(ROME_T_FSLIB_DEV_LIST, mptr);
    int count = 0;
    DOS_DEVICE *dev = dos.dev_list;
    FSLIB_DEVICE *curr_dev = p->devlist;

    /* get the offset */

    while (p->offset-- && dev)
    {
	dev = dev->next;
    }

    while (dev && p->count--)
    {
	curr_dev->name[0] = dev->devid;
	curr_dev->name[1] = '\0';
	curr_dev->attr.maxblk = dev->inf_valid ? dev->inf.totalsize : 0;

	curr_dev++;
	count++;

	dev = dev->next;
    }

    /* report the count of read devices */

    p->count = count;
    rome_reply(mptr);

}

void dos_m_fslib_link(ROME_MESSAGE *mptr)
{
    mptr->m_errno = ENOTSUP;
    rome_reply(mptr);

}

void dos_m_fslib_mkdir(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_MKDIR *p = RCASTP(ROME_T_FSLIB_MKDIR, mptr);
    char path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *dev;

    if ((dev = dos_setup(p->name, path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else
    {
	mptr->m_errno = dos_dir_mkdir(dev->fp, &dev->inf, path);
    }

    rome_reply(mptr);

}

void dos_m_fslib_readdir(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_READDIR *p = RCASTP(ROME_T_FSLIB_READDIR, mptr);
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    if  (of->dev)
    {
	mptr->m_errno = dos_dir_read_dir(of, p->dir, p->count, &p->count);
    }
    else
    {
	mptr->m_errno = ENODEV;
    }
    rome_reply(mptr);

}

void dos_m_fslib_rename(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_RENAME *p = RCASTP(ROME_T_FSLIB_RENAME, mptr);
    char from_path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *from_dev;

    if ((from_dev = dos_setup((char *)p->from, from_path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else
    {
	mptr->m_errno = dos_dir_rename_entry(from_dev->fp, &from_dev->inf,
					     from_path, p->to);
    }
    rome_reply(mptr);

}

void dos_m_fslib_rmdir(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_RMDIR *p = RCASTP(ROME_T_FSLIB_RMDIR, mptr);
    char path[MAX_FILE_NAME_LEN];
    DOS_DEVICE *dev;

    if ((dev = dos_setup(p->name, path)) == NULL)
    {
	mptr->m_errno = ENODEV;
    }
    else
    {
	mptr->m_errno = dos_dir_rmdir(dev->fp, &dev->inf, path);
    }

    rome_reply(mptr);

}

void dos_m_fslib_seek(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_SEEK *p = RCASTP(ROME_T_FSLIB_SEEK, mptr);
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    int newpos = 0;

    if (FSLIB_TYPE_FILE != of->type)
    {
	mptr->m_errno = ESPIPE;
	rome_reply(mptr);
	return;
    }

    switch (p->ptrname)
    {
	case SEEK_SET:
	    newpos = p->offset;
	    break;

	case SEEK_CUR:
	    newpos = of->fpos + p->offset;
	    break;

	case SEEK_END:
	    newpos = of->fend + p->offset;
	    break;

	default:
	    rome_kprintf("dos_fs_seek(): unknown ptrname %08x\n", p->ptrname);
    }

    if (newpos > of->fend)
    {
	mptr->m_errno = ESPIPE;
    }
    else
    {
	of->fpos = newpos;
    }

    rome_reply(mptr);

}

void dos_m_fslib_tell(ROME_MESSAGE *mptr)
{
    ROME_T_FSLIB_TELL *p = RCASTP(ROME_T_FSLIB_TELL, mptr);
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    if (FSLIB_TYPE_FILE != of->type)
    {
	mptr->m_errno = ESPIPE;
    }
    else
    {
	p->fpos = of->fpos;
    }

    rome_reply(mptr);

}

void dos_m_fslib_truncate(ROME_MESSAGE *mptr)
{
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    if  (of->dev == NULL)
    {
        mptr->m_errno = ENODEV;
    }
    else if (FSLIB_TYPE_FILE != of->type)
    {
	mptr->m_errno = ESPIPE;
    }
    else
    {
	mptr->m_errno = ENOTSUP;
    }

    rome_reply(mptr);

}

void dos_m_fetmblk(ROME_MESSAGE *mptr)
{
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    ushort next = 0;
    ROME_MESSAGE *mblk = &of->mblk;
    ushort blkidx;
    MSDOS_FS_INFO *inf;
    FILE *fp;
    int count = mptr->b_lim - mptr->b_wptr;
    int done = 0;

    if  (of->dev == NULL)
    {
	mptr->m_errno = ENODEV;
	rome_reply(mptr);
	return;
    }

    inf = &of->dev->inf;
    fp = of->dev->fp;

    if (mptr->opcode == ROME_M_GETMBLK)
    {
	mptr->b_base = rome_alloc(DOS_DEFAULT_MBLK_SIZE,
					   DOS_MEM_DEBUG, 0);
	rome_setup_mblk(mptr, mptr->b_base, DOS_DEFAULT_MBLK_SIZE);

	count = DOS_DEFAULT_MBLK_SIZE;
    }

    while (count)
    {
	if (of->fpos >= of->fend)
	{
	    if (done == 0)
	    {
		mptr->m_errno = EOF;
	    }
	    rome_reply(mptr);
	    return;
	}

	if (of->present && of->fpos >= of->mbstart && of->fpos < of->mbend)
	{
	    char c = *(of->mblk.b_rptr + of->fpos++ - of->mbstart);

	    /*  check for termination conditions  */

	    if (!(of->mode & FILE_BINARY_FLAG))
	    {
		if (c == '\r')
		{
		    ;
		}
		else if (c == '\n')
		{
		    *(mptr->b_wptr)++ = c;
		    *(mptr->b_wptr)++ = '\0';
		    break;
		}
		else
		{
		    *(mptr->b_wptr)++ = c;
		    count--;
		}
	    }
	    else
	    {
		*(mptr->b_wptr)++ = c;
		count--;
	    }
	    done++;
	    continue;
	}

	/*  need to get hold of another block -- return this one first */

	if (of->present)
	{
	    if (of->dirty)
	    {
		/* write it back */

		mblk->b_wptr = mblk->b_rptr + of->dev->inf.blksize;
		dos_fs_write_cluster(fp, inf, of->cluster, mblk);
	    }
	    else
	    {
		/* otherwise just give it back */
		rome_retmblk(mblk);
	    }

	    of->present = FALSE;
	    of->dirty = FALSE;
	}

	/* calculate index of new block */

	blkidx = of->fpos / of->dev->inf.blksize;

	/* Get the next cluster for reading (1 == EOF, 0 == free) */

	if ((of->blkidx + 1) == blkidx)
	{
	    /* if we have one in sequence, just get the link */
	    next = dos_fat_lookup_entry(of->cluster, inf);
	}
	else
	{
	    /* otherwise scan for the requested idx */

	    next = dos_fat_seek_entry(inf, of->start, blkidx);
	}

	if (2 > next)		/* block idx not in filescope */
	{
	    rome_kprintf("Read seek error\n");

	    mptr->m_errno = ESPIPE;
	    break;
	}

	dos_fs_read_cluster(fp, inf, next, mblk);

	of->present = TRUE;
	of->blkidx = blkidx;
	of->cluster = next;
	of->mbstart = blkidx * inf->blksize;
	of->mbend = (blkidx + 1) * inf->blksize;
	of->flim = (blkidx + 1) * inf->blksize;
    }

    rome_reply(mptr);

}

void dos_m_open(ROME_MESSAGE *mptr)
{
    ROME_T_OPEN *s = RCASTP(ROME_T_OPEN, mptr);
    int rc;
    char fn[MAX_FILE_NAME_LEN];
    ROME_URL *url = s->openurl;
    DOS_DEVICE *dev;
    DOS_OPENFILE *of;
    MSDOS_DIR_ENTRY entry;

    if ((dev = dos_setup(url->urlpath, fn)) == NULL)
    {
	mptr->m_errno = ENODEV;
	rome_reply(mptr);
	return;
    }

    /*  we don't link 'of' here in case the open fails  */

    of = rome_alloc(sizeof(DOS_OPENFILE), DOS_MEM_DEBUG, 1);
    strcpy(of->filename, fn);
    of->dev = dev;
    of->mode = s->mode;

    if  ('\0' == *fn)
    {
	/* This is the root directory -- the pointer fields are all NULL */

	of->type = FSLIB_TYPE_DIR;
    }
    else if (0 != (rc = dos_dir_get_entry(dev->fp, &dev->inf, fn, &entry)))
    {
	/*  file does not exist. Create it ? */

	if ((s->mode & FILE_WRITE_FLAG) &&
	    (0 == (rc = dos_dir_mkfile(dev->fp, &dev->inf, fn))))
	{
	    /*  reload the device entry for the new file  */

	    dos_dir_get_entry(dev->fp, &dev->inf, fn, &entry);
	    of->type = FSLIB_TYPE_FILE;
	    of->start = entry.start;
	    of->fend = entry.size;
	    of->cluster = of->start;
	}
	else
	{
	    rome_free(of);
	    mptr->m_errno = rc;
	    rome_reply(mptr);
	    return;
	}
    }
    else
    {
	/*  file exists, maybe a directory  */

	if (entry.attr & ATTR_DIR)
	{
	    of->type = FSLIB_TYPE_DIR;
	}
	else if ((s->mode & FILE_WRITE_FLAG) &&
		 (0 == (rc = dos_dir_rmfile(dev->fp, &dev->inf, fn))) &&
		 (0 == (rc = dos_dir_mkfile(dev->fp, &dev->inf, fn))))
	{
	    /*  synch & reload the device entry for the new file  */

	    of->type = FSLIB_TYPE_FILE;
	    dos_dir_get_entry(dev->fp, &dev->inf, fn, &entry);
	}
	else if (s->mode & FILE_READ_FLAG)
	{
	    /* update the last-accessed date */

	    of->type = FSLIB_TYPE_FILE;
	    entry.adate = dos_dir_mkdate();
	    dos_dir_write_entry(dev->fp, &dev->inf, fn, &entry);
	}
	else
	{
	    rome_free(of);
	    mptr->m_errno = rc;
	    rome_reply(mptr);
	    return;
	}

	of->start = entry.start;
	of->fend = entry.size;
	of->cluster = of->start;
    }

    /*  by this point we have a good open  */

    mptr->dest_context = of;
    of->link = dos.o_root;
    dos.o_root = of;
    mptr->m_errno = 0;
    rome_reply(mptr);

}

void dos_m_outmblk(ROME_MESSAGE *mptr)
{
    DOS_OPENFILE *of = (DOS_OPENFILE *)mptr->dest_context;
    ushort next = 0;
    ROME_MESSAGE *mblk = &of->mblk;
    ushort blkidx;
    MSDOS_FS_INFO *inf = &of->dev->inf;
    char c = '\0';
    FILE *fp = of->dev->fp;

    if  (of->dev == NULL)
    {
	if (ROME_M_PUTMBLK == mptr->opcode)
	{
	    rome_free(mptr->b_base);
	}
	mptr->m_errno = ENODEV;
        rome_reply(mptr);
	return;
    }

    while (mptr->b_rptr < mptr->b_wptr)
    {
	if (of->present && of->fpos < of->flim &&
	    of->fpos >= of->mbstart && of->fpos < of->mbend)
	{
	    if (c)
	    {
		*(of->mblk.b_rptr + of->fpos - of->mbstart) = c;
		c = '\0';
	    }
	    else
	    {
		if (*mptr->b_rptr == '\n' && !(of->mode & FILE_BINARY_FLAG))
		{
		    c = '\r';
		}

		*(of->mblk.b_rptr + of->fpos - of->mbstart) = *mptr->b_rptr++;
	    }
	    of->dirty = TRUE;
	    of->fpos++;

	    if (of->fpos >= of->fend)
	    {
		of->fend = of->fpos;
	    }
	    continue;
	}

	/* when we come here, we know, that the current buffer has to 
	 * be changed, because it's not the right block index.
	 * OR there's no buffer at all.
	 */

	if (of->present)
	{
	    if (of->dirty)
	    {
		mblk->b_wptr = mblk->b_rptr + of->dev->inf.blksize;
		dos_fs_write_cluster(fp, inf, of->cluster, mblk);
	    }
	    else
	    {
		rome_retmblk(mblk);
	    }

	    of->present = FALSE;
	    of->dirty = FALSE;
	}

	/* if we've moved to another block, reset blkidx */

	blkidx = of->fpos / of->dev->inf.blksize;

	if ((of->blkidx + 1) == blkidx)
	{
	    /* if we have one in sequence, just get the link */

	    next = dos_fat_lookup_entry(of->cluster, inf);
	}
	else
	    /* otherwise scan for the requested idx */
	{
	    next = dos_fat_seek_entry(inf, of->start, blkidx);
	}

	/* check the next cluster (1 EOF,0 not found ) */

	if (1 < next)
	{
	    /* we got a cluster somewhere within the file */

	    dos_fs_read_cluster(fp, inf, next, mblk);
	}
	else
	{
	    /* we eventually need a new block for a new cluster */
	    /* check, if the new block is the next in sequence */

	    if (of->fpos != of->fend)
	    {
		/* no -> illegal seek error */
		mptr->m_errno = ESPIPE;
		rome_reply(mptr);
		return;
	    }
	    else
	    {
		/* yes, append one */
		next = dos_fat_lookup_free(inf);
		if (0 == next)
		{
		    mptr->m_errno = ENOSPC;
		    rome_reply(mptr);
		    return;
		}
		else
		{
		    /* mark it as EOF */

		    if (0 != dos_fat_write_entry(inf, next, FAT_EOF))
		    {
			mptr->m_errno = EIO;
			rome_reply(mptr);
			return;
		    }
		}

		/* write the link to the FAT for current entry */

		if (0 != dos_fat_write_entry(inf, of->cluster, next))
		{
		    mptr->m_errno = EIO;
		    rome_reply(mptr);
		    return;
		}
	    }

	    rome_newmblk(fp, mblk, inf->blksize);
	    rome_await_message(mblk, 0);
	}			/* if 0 == next */

	/* we got the next buffer -> continue writing */

	of->present = TRUE;
	of->cluster = next;
	of->blkidx = blkidx;
	of->mbstart = blkidx * inf->blksize;
	of->mbend = (blkidx + 1) * inf->blksize;
	of->flim = (blkidx + 1) * inf->blksize;
    }

    if (ROME_M_PUTMBLK == mptr->opcode)
    {
	rome_free(mptr->b_base);
    }

    rome_reply(mptr);

}

/************************************************************************
 *									*
 *            Queue Handler						*
 *            =============						*
 *									*
 ************************************************************************/

int dos_qhandler(ROME_MESSAGE *mptr)
{
    switch (mptr->opcode)
    {
    case ROME_M_NEWMBLK :
    {
        int size = (int)mptr->b_wptr;
    
        if (DOS_DEFAULT_MBLK_SIZE > size)
        {
    	size = DOS_DEFAULT_MBLK_SIZE;
        }
    
        mptr->b_base = rome_alloc(size, DOS_MEM_DEBUG, 0);
        rome_setup_mblk(mptr, mptr->b_base, size);
        rome_auto_reply(mptr);
    }
    break;

    case ROME_M_PUTMBLK :
    {
        if  (mptr->b_wptr == mptr->b_rptr)
        {
	    rome_free(mptr->b_base);
	    rome_auto_reply(mptr);
        }
    }
    break;

    case ROME_M_RETMBLK :
    {
        rome_free(mptr->b_base);
        rome_auto_reply(mptr);
    }
    break;

    }
    return ROME_NOT_HANDLED;
}

static ROME_HANDLERS dos_routines[] =
{
    {ROME_M_CLOSE, dos_m_close},
    {ROME_M_COMMAND, dos_m_command},
    {ROME_R_EVENT, dos_r_event},
    {ROME_M_FSLIB_ATTR_GET, dos_m_fslib_attr_get},
    {ROME_M_FSLIB_ATTR_SET, dos_m_fslib_attr_set},
    {ROME_M_FSLIB_DELETE, dos_m_fslib_delete},
    {ROME_M_FSLIB_DEV_ATTR, dos_m_fslib_dev_attr},
    {ROME_M_FSLIB_DEV_LIST, dos_m_fslib_dev_list},
    {ROME_M_FSLIB_LINK, dos_m_fslib_link},
    {ROME_M_FSLIB_MKDIR, dos_m_fslib_mkdir},
    {ROME_M_FSLIB_READDIR, dos_m_fslib_readdir},
    {ROME_M_FSLIB_RENAME, dos_m_fslib_rename},
    {ROME_M_FSLIB_RMDIR, dos_m_fslib_rmdir},
    {ROME_M_FSLIB_SEEK, dos_m_fslib_seek},
    {ROME_M_FSLIB_TELL, dos_m_fslib_tell},
    {ROME_M_FSLIB_TRUNCATE, dos_m_fslib_truncate},
    {ROME_M_FETMBLK, dos_m_fetmblk},
    {ROME_M_GETMBLK, dos_m_fetmblk},
    {ROME_M_OPEN, dos_m_open},
    {ROME_M_OUTMBLK, dos_m_outmblk},
    {ROME_M_PUTMBLK, dos_m_outmblk} 
};

/************************************************************************
 *									*
 *            Main Process						*
 *            ============						*
 *									*
 ************************************************************************/

void dos_process(void)
{
    while (1==1)
    {
        rome_generic_handler(rome_await_message(0, 0),
            dos_routines, 21);
    }
}
